"""
Generate subfigures for Figure 9, which is a stochastic, spatial simulation

All the subfigures can be generated by calling the following command with an experiment_id:
    `python bioscrape_cobra/generate_figure.py experiment_id --all`

To access the saved experiments on the mongoDB database, install vivarium scripts with:
    `pip install vivarium-scripts`
and run the commands:
    `python -m scripts.access_db list`
    `python -m scripts.access_db info [experiment_id]`
"""

import os
import argparse
import copy

from vivarium.core.emitter import (
    data_from_database,
    DatabaseEmitter,
    remove_units, deserialize_value)
from vivarium.plots.simulation_output import set_axes
from vivarium.plots.simulation_output import _save_fig_to_dir as save_fig_to_dir
from vivarium.plots.agents_multigen import plot_agents_multigen
from vivarium_multibody.plots.snapshots import (
    format_snapshot_data, make_tags_figure, get_agent_colors)
from bioscrape_cobra.plot import (
    plot_fields_tags, plot_fields_snapshots)
from bioscrape_cobra.bioscrape_cobra_stochastic import (
    GLUCOSE_EXTERNAL, LACTOSE_EXTERNAL)


MULTIGEN_PLOT_CONFIG = {
    'include_paths': [
        ('species', 'rna_M'),
        ('species', 'protein_betaGal'),
        ('species', 'protein_Lactose_Permease'),
        ('species', 'Glucose_external'),
        ('species', 'Lactose_external'),
        # ('flux_bounds', 'EX_glc__D_e'),
        # ('flux_bounds', 'EX_lac__D_e'),
        ('boundary', 'volume'),
        ('boundary', 'mass'),
    ],
    'store_order': ('species', 'flux_bounds', 'boundary'),
    'titles_map': {
        ('species', 'rna_M'): 'lac operon RNA (counts)',
        ('species', 'protein_betaGal'): r'$\beta$-galactosidase (counts)',
        ('species', 'protein_Lactose_Permease'): 'lactose permease (counts)',
        ('species', 'Glucose_external'): 'local external glucose (counts)',
        ('species', 'Lactose_external'): 'local external lactose (counts)',
        # ('flux_bounds', 'EX_glc__D_e'): 'glucose flux bound (mmol/L/s)',
        # ('flux_bounds', 'EX_lac__D_e'): 'lactose flux bound (mmol/L/s)',
        ('boundary', 'volume'): 'volume (fL)',
    },
    'remove_zeros': False,
    'column_width': 5,
    'row_height': 1.5,
    'title_on_y_axis': False,
    'stack_column': True,
    'tick_label_size': 10,
    'title_size': 10,
    'sci_notation': 3,
    'linewidth': 2,
}

YLABEL_SIZE = 48

def access(experiment_id):
    """ Access an experiment from the simulation database """

    # mongo client
    config = {
        'host': '{}:{}'.format('localhost', 27017),
        'database': 'simulations'}
    emitter = DatabaseEmitter(config)
    db = emitter.db

    # access experiment from database
    data, experiment_config = data_from_database(experiment_id, db)

    # get the bounds
    bounds = experiment_config['processes']['diffusion']['bounds']

    # reformat data
    deserialized = deserialize_value(data)
    output = remove_units(deserialized)

    return output, bounds


def plot_fields_fig(
        output,
        bounds,
        out_dir
):
    """ plot snapshots of the glucose and lactose fields """

    new_output = copy.deepcopy(output)
    seconds_list = list(new_output.keys())
    for seconds in seconds_list:
        hours = round(seconds/60/60, 2)
        new_output[hours] = new_output.pop(seconds)

    # plot snapshots fields
    fig_snapshots = plot_fields_snapshots(
        new_output,
        bounds=bounds,
        include_fields=[GLUCOSE_EXTERNAL, LACTOSE_EXTERNAL],
        colorbar_decimals=1,
        scale_bar_color='white',
        agent_fill_color='k',
        show_timeline=True,
        time_unit='hr')

    # alter figure labels and save
    axes = fig_snapshots.get_axes()
    for axis in axes:
        axis.set_title('')
        ylabel = axis.get_ylabel()
        if ylabel == 'Glucose_external':
            axis.set_ylabel('')
            axis.set_title('external glucose concentrations (mmol)', fontsize=YLABEL_SIZE, pad=15, x=2.3)
        if ylabel == 'Lactose_external':
            axis.set_ylabel('')
            axis.set_title('external lactose concentrations (mmol)', fontsize=YLABEL_SIZE, pad=15, x=2.3)

    save_fig_to_dir(
        fig_snapshots,
        out_dir=out_dir,
        filename='field_snapshots.pdf')


def plot_phylogeny_fig(
        output,
        bounds,
        agent_colors=None,
        out_dir='out'
):
    """ plot snapshots of the colony, with phylogeny coloring """

    fig_phylogeny = plot_fields_snapshots(
        output,
        bounds=bounds,
        skip_fields=[GLUCOSE_EXTERNAL, LACTOSE_EXTERNAL],
        agent_colors=agent_colors,
        colorbar_decimals=1,
        show_timeline=False,
        filename='phylogeny_snapshots.pdf',
        out_dir=out_dir)


def plot_tags_fig(
        output,
        bounds,
        out_dir
):
    """ plot snapshots of the lactose permease tags """

    fig_tags = plot_fields_tags(
        output,
        bounds=bounds,
        convert_to_concs=False,
        tagged_molecules=[('species', 'protein_Lactose_Permease',)],
        colorbar_decimals=1)

    # alter figure and save
    axes = fig_tags.get_axes()
    for axis in axes:
        axis.set_title('')
        ylabel = axis.get_ylabel()
        if ylabel:
            axis.set_ylabel('internal\nlactose\npermease', fontsize=YLABEL_SIZE)
    save_fig_to_dir(
        fig_tags,
        out_dir=out_dir,
        filename='tags_snapshots.pdf')


def plot_multigen_fig(
        output,
        agent_colors=None,
        out_dir='out'
):
    """ plot multi-generational timeseries over the entire simulation """

    plot_config = copy.deepcopy(MULTIGEN_PLOT_CONFIG)
    plot_config['agent_colors'] = agent_colors

    # plot multigen
    multigen_fig = plot_agents_multigen(output, plot_config)

    # add colony mass to multigen_fig
    colony_mass, time_vec = calculate_colony_mass(output)

    # convert to hours
    time_vec = [time/3600 for time in time_vec]

    # get mass axis to replace with colony mass
    allaxes = multigen_fig.get_axes()
    ax = None
    for axis in allaxes:
        if axis.get_title() == 'boundary \nmass':
            ax = axis

    # replace with colony mass
    ax.clear()
    set_axes(ax, True, sci_notation=MULTIGEN_PLOT_CONFIG['sci_notation'])
    ax.plot(time_vec, colony_mass,
            linewidth=MULTIGEN_PLOT_CONFIG['linewidth'],
            color='darkslategray')
    ax.set_xlim([time_vec[0], time_vec[-1]])
    ax.set_title('total colony mass (fg)',
                 rotation=0,
                 fontsize=MULTIGEN_PLOT_CONFIG['title_size'])
    ax.set_xlabel('time (hr)')
    ax.spines['bottom'].set_position(('axes', -0.2))

    # save
    save_fig_to_dir(multigen_fig, 'spatial_multigen.pdf', out_dir)


def plot_single_tags(
        agents,
        bounds,
        out_dir
):
    """ plot a final snapshot with several different tags """

    # make individual tag plots
    time_vec = list(agents.keys())
    snapshot_times = [time_vec[-1]]
    time_indices = [time_vec.index(time) for time in snapshot_times]

    ############
    # glc flux #
    ############
    fig_tags = make_tags_figure(
        agents=agents, bounds=bounds, n_snapshots=1,
        time_indices=time_indices, snapshot_times=snapshot_times,
        background_color='white', scale_bar_length=False, show_timeline=False, convert_to_concs=False,
        tagged_molecules=[('flux_bounds', 'EX_glc__D_e')])

    # alter figure and save
    axes = fig_tags.get_axes()
    axes[0].set_title('glucose flux\n(mmol/L/s)', fontsize=YLABEL_SIZE, pad=15)
    axes[0].get_yaxis().set_visible(False)
    save_fig_to_dir(fig_tags, out_dir=out_dir, filename='tag_EX_glc__D_e.pdf')


    ############
    # lac flux #
    ############
    fig_tags = make_tags_figure(
        agents=agents, bounds=bounds, n_snapshots=1,
        time_indices=time_indices, snapshot_times=snapshot_times,
        background_color='white', scale_bar_length=False, show_timeline=False, convert_to_concs=False,
        tagged_molecules=[('flux_bounds', 'EX_lac__D_e')])
    # alter figure and save
    axes = fig_tags.get_axes()
    axes[0].set_title('lactose flux\n(mmol/L/s)', fontsize=YLABEL_SIZE, pad=15)
    axes[0].get_yaxis().set_visible(False)
    save_fig_to_dir(fig_tags, out_dir=out_dir, filename='tag_EX_lac__D_e.pdf')

    ################
    # lac Permease #
    ################
    fig_tags = make_tags_figure(
        agents=agents, bounds=bounds, n_snapshots=1,
        time_indices=time_indices, snapshot_times=snapshot_times, colorbar_decimals=0,
        background_color='white', scale_bar_length=False, show_timeline=False, convert_to_concs=False,
        tagged_molecules=[('species', 'protein_Lactose_Permease')])
    # alter figure and save
    axes = fig_tags.get_axes()
    axes[0].set_title('lactose permease\n(counts)', fontsize=YLABEL_SIZE, pad=15)
    axes[0].get_yaxis().set_visible(False)
    save_fig_to_dir(fig_tags, out_dir=out_dir, filename='tag_protein_Lactose_Permease.pdf')


    ###############
    # growth rate #
    ###############
    agents = add_growth_rate_to_agents(agents)
    fig_tags = make_tags_figure(
        agents=agents, bounds=bounds, n_snapshots=1,
        time_indices=time_indices, snapshot_times=snapshot_times,
        background_color='white', scale_bar_length=False, show_timeline=False, convert_to_concs=False,
        tagged_molecules=[('boundary', 'growth_rate')])
    # alter figure and save
    axes = fig_tags.get_axes()
    axes[0].set_title('growth rate\n(fg/s)', fontsize=YLABEL_SIZE, pad=15)
    axes[0].get_yaxis().set_visible(False)
    save_fig_to_dir(fig_tags, out_dir=out_dir, filename='tag_growth_rate.pdf')


def add_growth_rate_to_agents(agents):
    """ calculate individual agent growth rates and add it to the data """

    time_vec = list(agents.keys())

    # initial state at 0
    for agent_id, state in agents[time_vec[0]].items():
        state['boundary']['growth_rate'] = 0

    for t0, t1 in zip(time_vec[:-1], time_vec[1:]):
        agents0 = agents[t0]
        agents1 = agents[t1]
        for agent_id, state1 in agents1.items():
            if agent_id in agents0:
                mass0 = agents0[agent_id]['boundary']['mass']
                mass1 = state1['boundary']['mass']
                dm = mass1 - mass0
                dt = t1 - t0
                growth_rate = dm/dt
                state1['boundary']['growth_rate'] = growth_rate
    return agents


def calculate_colony_mass(output):
    """ calculate total colony mass """

    colony_mass = []
    time_vec = list(output.keys())
    for time, state in output.items():
        mass = 0.0
        for agent_id, agent_state in state['agents'].items():
            mass += agent_state['boundary']['mass']
        colony_mass.append(mass)
    return colony_mass, time_vec


def main():
    parser = argparse.ArgumentParser(description='access data from db')
    parser.add_argument('experiment_id', type=str, default=False)
    parser.add_argument('--multigen', '-1', action='store_true', default=False)
    parser.add_argument('--fields', '-2', action='store_true', default=False)
    parser.add_argument('--tags', '-3', action='store_true', default=False)
    parser.add_argument('--single_tags', '-4', action='store_true', default=False)
    parser.add_argument('--all', '-a', action='store_true', default=False)
    args = parser.parse_args()
    experiment_id = args.experiment_id


    # make a directory for the figures
    out_dir = f'out/analyze/{experiment_id}'
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # retrieve the data
    output, bounds = access(experiment_id)
    del output[0.0]
    for t in list(output.keys()):
        if t > 43200: # cutoff at 12 hours
            del output[t]

    agents, fields = format_snapshot_data(output)

    # run the plot functions
    if args.multigen or args.all:
        agent_colors = get_agent_colors(agents)
        plot_phylogeny_fig(output, bounds, agent_colors, out_dir)
        plot_multigen_fig(output, agent_colors, out_dir)

    if args.fields or args.all:
        plot_fields_fig(output, bounds, out_dir)

    if args.tags or args.all:
        plot_tags_fig(output, bounds, out_dir)

    if args.single_tags or args.all:
        plot_single_tags(agents, bounds, out_dir)


if __name__ == '__main__':
    main()



import matplotlib.pyplot as plt

# plotting
from vivarium.plots.simulation_output import plot_variables
from vivarium.plots.agents_multigen import plot_agents_multigen
from vivarium_multibody.plots.snapshots import format_snapshot_data, plot_snapshots, plot_tags
import copy

def plot_single(
        output,
        variables=None,
        config=None,
        out_dir=None,
        filename=None
):
    variables = variables or []
    config = config or {}
    variables_plot_config = {
        'out_dir': out_dir,
        'filename': filename,
        'row_height': 1.2,
        'row_padding': 0.6,
        'column_width': 4,
        'default_color': 'tab:gray',
        'linewidth': 2.0,
        'sci_notation': 3,
        'variables': variables}
    variables_plot_config.update(config)
    fig = plot_variables(
        output, **variables_plot_config)
    return fig

def plot_multigen(
        output,
        variables=None,
        out_dir=None,
        filename=None,
        **kwargs
):
    if filename and not out_dir:
        out_dir = 'out'

    plot_settings = {
        'include_paths': variables,
        'skip_paths': [
            ('internal_counts',),
            ('cobra_external',),
        ],
        'remove_zeros': False,
        'column_width': 8,
        'row_height': 1.5,
        'title_on_y_axis': True,
        'stack_column': True,
        'tick_label_size': 10,
        'title_size': 10,
        **kwargs
        }

    fig = plot_agents_multigen(
        output,
        plot_settings,
        out_dir,
        filename)
    return fig


def plot_fields_snapshots(
        output,
        bounds=None,
        include_fields=None,
        n_snapshots=4,
        colorbar_decimals=4,
        scale_bar_length=5,
        phylogeny_colors=False,
        out_dir=None,
        filename=None,
        **kwargs):

    agents, fields = format_snapshot_data(output)

    fig1 = plot_snapshots(
        bounds=bounds,
        agents=agents,
        fields=fields,
        phylogeny_names=phylogeny_colors,
        n_snapshots=n_snapshots,
        scale_bar_length=scale_bar_length,
        colorbar_decimals=colorbar_decimals,
        include_fields=include_fields,
        out_dir=out_dir,
        filename=filename,
        **kwargs)

    return fig1

# plotting
tags_dict = {
    'glc__D_e': 'tab:cyan',
    'co2_e': 'tab:orange',
    'h2o_c': 'tab:cyan',
    'atp_c': 'tab:orange',
    'asp__L_c': 'tab:green'
}
def plot_fields_tags(
        output,
        bounds=None,
        tagged_molecules=None,
        n_snapshots=4,
        colorbar_decimals=4,
        out_dir=None,
        filename=None,
        **kwargs):

    fig2 = plot_tags(
        data=output,
        bounds=bounds,
        tagged_molecules=tagged_molecules,
        n_snapshots=n_snapshots,
        background_color='w',
        scale_bar_length=5,
        colorbar_decimals=colorbar_decimals,
        out_dir=out_dir,
        filename=('tags_' + filename) if filename else None,
        **kwargs
    )
    
    return fig2




def move_to_end(data, d):
    for key in d.keys():
        if key in data:
            data[key] = data.pop(key)
    return data


def plot_metabolism(data, tags=tags_dict):
    # initialize subplots

    n_rows = 3
    n_cols = 1
    fig = plt.figure(figsize=(n_cols * 6, n_rows * 4/3))
    grid = plt.GridSpec(n_rows, n_cols)

    time_vec = data['time']

    # external
    ax = fig.add_subplot(grid[0, 0])
    external = move_to_end(data['external'], tags)
    for mol_id, series in external.items():
        if sum(series) != 0.0:
            ax.plot(
                time_vec,
                series,
                label=mol_id if mol_id in tags else None,
                color=tags.get(mol_id, 'tab:gray'),
            )
    ax.set_title(f'external ({len(external.keys())} species)')
    ax.set_ylabel('concentrations (mmol)')
    ax.set_yscale('log')

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.tick_params(bottom=False, labelbottom=False)
    ax.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))

    # internal
    ax = fig.add_subplot(grid[1, 0])
    internal = move_to_end(data['internal_counts'], tags)
    for mol_id, series in internal.items():
        if sum(series) != 0.0:
            ax.plot(
                time_vec,
                series,
                label=mol_id if mol_id in tags else None,
                color=tags.get(mol_id, 'tab:gray'),
            )
    ax.set_title(f'internal ({len(internal.keys())} species)')
    ax.set_ylabel('counts')
    ax.set_yscale('log')

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.tick_params(bottom=False, labelbottom=False)
    ax.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))

    # mass
    ax = fig.add_subplot(grid[2, 0])
    ax.plot(
        time_vec,
        data['global'][('mass', 'femtogram')],
        color='tab:blue', label="biomass (fg)"
    )
    ax.set_ylabel('cell mass')
    ax.set_title('global')
    ax.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))

    #Set the x label for the last axis, whichever it is
    ax.set_xlabel('time (sec)')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    fig.tight_layout()
    return fig


# Below are different configs for the topology plots
custom_widths = {
    'cobra':8,
    'bioscrape':8,
    'division':8,
    'diffusion':8,
    'multibody':8
}

embedded_custom_widths = {"agents\n0\n"+k:custom_widths[k] for k in custom_widths if k not in ["diffusion", "multibody"]}
embedded_custom_widths['diffusion'] = custom_widths['diffusion']
embedded_custom_widths['multibody'] = custom_widths['multibody']

u = 1.0
r0 = 0
r1 = 1.5*u
cobra_topology_config = {
    'graph_format': 'hierarchy',
    'dashed_edges': True,
    'show_ports': False,
    'remove_nodes':[
        'exchanges', 'field_deriver', 'global\nlocation', 'dimensions',
    ],
    'process_colors':{
        'cobra':'tab:orange',
        'mass\nderiver':'tab:orange',
        'volume\nderiver':'tab:orange',
    },
    'store_colors':{
        'internal_counts':'tab:orange',
        'external':'tab:orange',
        'exchanges':'tab:orange',
        'reactions':'tab:orange',
        'flux_bounds':'tab:orange',
        'global':'tab:brown',
    },
    'node_labels':{
        'mass_deriver':'mass\nderiver',
        'volume_deriver':'volume\nderiver',
        'internal_counts':'internal\ncounts',
        'flux_bounds':'flux\nbounds',
    },
    'coordinates':{
        
        'cobra':(-1*u, r1),
        'mass_deriver':(0*u, r1),
        'volume_deriver':(1*u, r1),
        
        'reactions':(-2*u, r0),
        'internal_counts':(-1*u, r0),
        'external':(0*u, r0),
        'flux_bounds':(1*u, r0),
        'global':(2*u, r0)
    },
    'custom_widths':custom_widths
}

bioscrape_topology_config = {
    'dashed_edges':True,
    'show_ports':False,
    'process_colors':{
        'bioscrape':'tab:green',
    },
    'store_colors':{
        'species':'tab:green',
        'delta_species':'tab:green',
        'rates':'tab:green',
        'globals':'tab:brown'
    },
    'node_labels':{
        'delta_species':'delta\nspecies'
    },
    'custom_widths':custom_widths
}


u = 0.9
config_grow_divide_topology_i = {
    'graph_format': 'hierarchy',
    'dashed_edges':True,
    'show_ports':False,
    'remove_nodes':[
        'agents\n0\nrates\ngrowth_rate', 'agents\n0\nrates\ngrowth_noise', 'dimensions'
    ],
    'process_colors':{
        'agents\n0\ndivide_condition':'tab:cyan',
        'agents\n0\ndivision':'tab:cyan',
        'agents\n0\nglobals_deriver':'tab:gray',
        'agents\n0\ngrowth':'tab:gray',
    },
    'store_colors':{
        'agents':'tab:cyan',
        'agents\n0':'tab:cyan',
        'agents\n0\nrates':'tab:gray',
        'agents\n0\nboundary':'tab:brown',
        'agents\n0\nboundary\nmass':'tab:brown',
        'agents\n0\nboundary\ndivide':'tab:brown',
    },
    'coordinates':{
        'agents\n0\ndivision':(u, -3*u),
        'agents\n0\ndivide_condition':(2*u, -3*u),
        
        'agents':(0,  u),
        'agents\n0':(0, 0),
        'agents\n0\nrates':(-1*u, -u),
        'agents\n0\nboundary':(0*u, -u),
        'agents\n0\nboundary\nmass':(1*u, -1.5*u),
        'agents\n0\nboundary\ndivide':(2*u, -1.5*u),
        
        'agents\n0\ngrowth':(-1*u, -3*u),
        'agents\n0\nglobals_deriver':(0*u, -3*u)
    },
    'node_labels':{
        'agents\n0':"agent\n0",
        'agents\n0\nglobals_deriver':'agent 0\nglobals\nderiver',
        'agents\n0\ndivide_condition':'agent 0\ndivide\ncondition',
        'agents\n0\ndivision':'agent 0\ndivision',
        'agents\n0\nrates':'agent 0\nrates',
        'agents\n0\ngrowth':'agent 0\ngrowth',
        'agents\n0\ngloabls\nderiver':'agent 0\nglobals\nderiver',
        'agents\n0\nboundary':'agent 0\nboundary',
        'agents\n0\nboundary\nmass':'agent 0\nboundary\nmass',
        'agents\n0\nboundary\ndivide':'agent 0\nboundary\ndivide',
    },
    'custom_widths':embedded_custom_widths
}

embedded_custom_widths_f = {}
embedded_custom_widths_00 = {"agents\n00\n"+k:custom_widths[k] for k in custom_widths if k not in ["diffusion", "multibody"]}
embedded_custom_widths_01 = {"agents\n01\n"+k:custom_widths[k] for k in custom_widths if k not in ["diffusion", "multibody"]}
embedded_custom_widths_f.update(embedded_custom_widths_00)
embedded_custom_widths_f.update(embedded_custom_widths_01)

config_grow_divide_topology_f = {
    'graph_format': 'hierarchy',
    'dashed_edges':True,
    'show_ports':False,
    'remove_nodes':[
        'agents\n00\nrates\ngrowth_rate', 'agents\n00\nrates\ngrowth_noise',
        'agents\n01\nrates\ngrowth_rate', 'agents\n01\nrates\ngrowth_noise',
        'dimensions'
    ],
    'process_colors':{
        'agents\n00\ndivide_condition':'tab:cyan',
        'agents\n00\ndivision':'tab:cyan',
        'agents\n00\nglobals_deriver':'tab:gray',
        'agents\n00\ngrowth':'tab:gray',
        
        'agents\n01\ndivide_condition':'tab:cyan',
        'agents\n01\ndivision':'tab:cyan',
        'agents\n01\nglobals_deriver':'tab:gray',
        'agents\n01\ngrowth':'tab:gray',
    },
    'store_colors':{
        'agents':'tab:cyan',
        'agents\n00':'tab:cyan',
        'agents\n00\nrates':'tab:gray',
        'agents\n00\nboundary':'tab:brown',
        'agents\n00\nboundary\nmass':'tab:brown',
        'agents\n00\nboundary\ndivide':'tab:brown',
        
        'agents\n01':'tab:cyan',
        'agents\n01\nrates':'tab:gray',
        'agents\n01\nboundary':'tab:brown',
        'agents\n01\nboundary\nmass':'tab:brown',
        'agents\n01\nboundary\ndivide':'tab:brown',
    },
    'coordinates':{
        
        'agents':(3*u,  u),
        
        'agents\n00\ndivision':(u, -3*u),
        'agents\n00\ndivide_condition':(2*u,  -3*u),
        'agents\n00':(0, 0),
        'agents\n00\nrates':(-1*u, -u),
        'agents\n00\nboundary':(0*u, -u),
        'agents\n00\nboundary\nmass':(1*u, -1.5*u),
        'agents\n00\nboundary\ndivide':(2*u, -1.5*u),
        'agents\n00\ngrowth':(-1*u, -3*u),
        'agents\n00\nglobals_deriver':(0*u, -3*u),
        
        'agents\n01\ndivision':(5.5*u, -3*u),
        'agents\n01\ndivide_condition':(6.5*u, -3*u),
        'agents\n01':(4.5*u, 0),
        'agents\n01\nrates':(3.5*u, -u),
        'agents\n01\nboundary':(4.5*u, -u),
        'agents\n01\nboundary\nmass':(5.5*u, -1.5*u),
        'agents\n01\nboundary\ndivide':(6.5*u, -1.5*u),
        'agents\n01\ngrowth':(3.5*u, -3*u),
        'agents\n01\nglobals_deriver':(4.5*u, -3*u)

    },
    'node_labels':{
        'agents\n00':"agent\n0",
        'agents\n00\nglobals_deriver':'agent 0\nglobals\nderiver',
        'agents\n00\ndivide_condition':'agent 0\ndivide\ncondition',
        'agents\n00\ndivision':'agent 0\ndivision',
        'agents\n00\nrates':'agent 0\nrates',
        'agents\n00\ngrowth':'agent 0\ngrowth',
        'agents\n00\ngloabls\nderiver':'agent 0\nglobals\nderiver',
        'agents\n00\nboundary':'agent 0\nboundary',
        'agents\n00\nboundary\nmass':'agent 0\nboundary\nmass',
        'agents\n00\nboundary\ndivide':'agent 0\nboundary\ndivide',
        
        'agents\n01':"agent\n1",
        'agents\n01\nglobals_deriver':'agent 1\nglobals\nderiver',
        'agents\n01\ndivide_condition':'agent 1\ndivide\ncondition',
        'agents\n01\ndivision':'agent 1\ndivision',
        'agents\n01\nrates':'agent 1\nrates',
        'agents\n01\ngrowth':'agent 1\ngrowth',
        'agents\n01\ngloabls\nderiver':'agent 1\nglobals\nderiver',
        'agents\n01\nboundary':'agent 1\nboundary',
        'agents\n01\nboundary\nmass':'agent 1\nboundary\nmass',
        'agents\n01\nboundary\ndivide':'agent 1\nboundary\ndivide',
    },
    'custom_widths':embedded_custom_widths_f
}

config_diffusion_topology = {
    'dashed_edges':True,
    'show_ports':False,
    'process_colors':{
        'diffusion_field':'tab:blue',
    },
    'store_colors':{
        'fields':'tab:blue',
        'agents':'tab:cyan',
        'dimensions':'tab:blue'
    },
    'node_labels':{
        'diffusion_field':'diffusion'
    },
    'custom_widths':{
        'diffusion_field':8
        }
}

config_lattice_topology = {
    'dashed_edges':True,
    'show_ports':False,
    'process_colors':{
        'multibody':'tab:blue',
        'diffusion':'tab:blue',
    },
    'store_colors':{
        'fields':'tab:blue',
        'agents':'tab:cyan',
        'dimensions':'tab:blue'
    },
    'coordinates': {
        'diffusion': (1.5, 1),
        'multibody': (2.5, 1),
    },
    'custom_widths':custom_widths
}

u = .8
config_grow_divide_lattice_topology = {
    'graph_format': 'hierarchy',
    'dashed_edges': True,
    'show_ports': False,
    'remove_nodes':[
        'agents\n0\nrates\ngrowth_rate', 'agents\n0\nrates\ngrowth_noise',
        # 'dimensions'
    ],
    'process_colors':{
        'agents\n0\ndivide_condition':'tab:cyan',
        'agents\n0\ndivision':'tab:cyan',
        'multibody':'tab:blue',
        'diffusion':'tab:blue',
        'agents\n0\nglobals_deriver':'tab:gray',
        'agents\n0\ngrowth':'tab:gray',
    },
    'store_colors':{
        'agents':'tab:cyan',
        'agents\n0':'tab:cyan',
        'agents\n0\nrates':'tab:gray',
        'agents\n0\nboundary':'tab:brown',
        'agents\n0\nboundary\nmass':'tab:brown',
        'agents\n0\nboundary\nexternal':'tab:brown',
        'agents\n0\nboundary\ndivide':'tab:brown',
    },
    'coordinates':{
        'agents\n0\ndivision':(1.5*u, -3*u),
        'agents\n0\ndivide_condition':(2.5*u, -3*u),
        'multibody':(2*u, 2*u),
        'diffusion':(3*u, 2*u),
        
        'agents':(0,  u),
        'agents\n0':(0, 0),
        'agents\n0\nrates':(-1*u, -u),
        'agents\n0\nboundary':(0*u, -u),
        'agents\n0\nboundary\nexternal':(1*u, -1.5*u),
        'agents\n0\nboundary\nmass':(2*u, -1.5*u),
        'agents\n0\nboundary\ndivide':(3*u, -1.5*u),
        
        'fields':(2.5*u, 1*u),
        'dimensions':(3.5*u, 1*u),
        'agents\n0\ngrowth':(-.5*u, -3*u),
        'agents\n0\nglobals_deriver':(.5*u, -3*u)
    },
    'node_labels':{
        'agents\n0':"agent\nn",
        'agents\n0\nglobals_deriver':'globals\nderiver',
        'agents\n0\ndivide_condition':'divide\ncondition',
        'agents\n0\ndivision':'division',
        'agents\n0\nrates':'rates',
        'agents\n0\ngrowth':'growth',
        'agents\n0\ngloabls\nderiver':'globals\nderiver',
        'agents\n0\nboundary':'boundary',
        'agents\n0\nboundary\nexternal':'boundary\nexternal',
        'agents\n0\nboundary\nmass':'boundary\nmass',
        'agents\n0\nboundary\ndivide':'boundary\ndivide',
    },
    'custom_widths':embedded_custom_widths
}


su = .8
RS = 0 #store row
RP = -1.8*su #process row

# node coordinates
agent_coordinates = {       
        # cobra
        'cobra':(-4*su, RP),
        'mass_deriver':(-3*su, RP),
        'volume_deriver':(-2*su, RP),
        'reactions':(-4*su, RS),
        'internal_counts':(-3*su, RS),
        'cobra_external':(-2*su, RS),
        'flux_bounds':(-1*su, RS),
        'fields':(4*su, RS-.5*su),

        # bioscrape
        'flux_adaptor':(-1*su, RP),
        'bioscrape':(0*su, RP),
        'dilution_rate_adaptor':(1*su, RP),
        'biomass_adaptor':(2*su, RP),
        'delta_counts_to_concs':(1*su, RP),
        'species':(0, RS),
        'rates':(su, RS),
        'delta_species':(2*su, RS),
        'delta_concs':(3*su, RS),
        
        # boundary/division
        'local_field':(5*su, RP),
        'field_counts_deriver':(6*su, RP),
        'division':(3*su, RP),
        'divide_condition':(4*su, RP),
        'boundary':(3*su, RS),
        'boundary\nexternal':(5*su, RS+.5*su),
        'boundary\nlocation':(4*su, RS+.5*su),
        'boundary\nexchanges':(5.5*su, RS-su),
        'boundary\nmass':(4*su, RS-.5*su),
        'boundary\ndivide':(5*su, RS-.5*su)}
global_coordinates = {
        'agents':(0, RS+2*su),
        'agents\n0':(0, RS+su),
        'multibody':(5.5*su, RS+2*su),
        'diffusion':(6.5*su, RS+2*su),
        'dimensions': (6*su, RS+su),
        'fields':(7*su, RS+su),
        # 'fields': (2.5 * su, 1 * su),
}
embedded_coordinates = {
    'agents\n0\n'+node_id: coord 
    for node_id, coord in agent_coordinates.items()}
embedded_coordinates.update(global_coordinates)



# Bioscrape Cobra Config
# node labels
agent_node_labels = {
        'flux_bounds':'flux\nbounds',
        'internal_counts':'internal\ncounts',
        'cobra_external':'cobra\nexternal',
        'hidden_exchanges':'hidden\nexchanges',
        'flux_adaptor':'flux\nadaptor',
        'dilution_rate_adaptor':'dilution rate\nadaptor',
        'biomass_adaptor':'biomass\nadaptor',
        'mass_deriver':'mass\nderiver',
        'volume_adaptor':'volume\nadaptor',
        'delta_species:':'delta\nspecies',
        'mass_deriver':'mass\nderiver',
        'volume_deriver':'volume\nderiver',
        'local_field': 'local\nfield',
        'boundary\nno_units':'boundary\n(no units)',
        'strip_units':'strip\nunits',
        'delta_concentrations':'delta\nconcs',
        'delta_counts_to_concs':'delta counts\nto\nconcs',
        'delta_species':'delta\nspecies',
        'field_counts_deriver':"field\ncounts\nderiver",

        'cobra':'cobra',
        'bioscrape':'bioscrape',
        'flux_adaptor':'flux\nadaptor',
        'dilution_rate_adaptor':'dilution rate\nadaptor',
        'clock':'clock',
        'mass_deriver':'mass\nderiver',
        'volume_deriver':'volume\nderiver',
        'biomass_adaptor':'biomass\nadaptor',
        'strip_units':'strip\nunits',
        'local_field':'local\nfield',
        'divide_condition':'divide\ncondition',
        'division':'division',
        'delta_counts_to_concs':'delta counts\nto\nconcs',
        'field_counts_deriver':'field\ncounts\nderiver',
        
        'species':'species',
        'internal_counts':"internal\ncounts",
        'delta_species':'delta\nspecies',
        'reactions':'reactions',
        'internal_counts':'internal\ncounts',
        'rates':'rates',
        'boundary':'boundary',
        'cobra_external':'cobra\nexternal',
        'hidden_exchanges':'hidden\nexchanges',
        'flux_bounds':'flux\nbounds',
        'boundary\ntime':'boundary\ntime',
        'boundary\nexternal':'boundary\nexternal',
        'boundary\ndivide':'boundary\ndivide',
        'boundary\nmass':'boundary\nmass',
        'boundary\nexchanges':'boundary\nexchanges',
        'boundary\nlocation':'boundary\nlocation',
        'boundary\nno_units':'boundary\n(no units)'
    }
global_node_labels = {
        'agents\n0':'agent\nn'}
embedded_node_labels = {
    'agents\n0\n'+node_id: label 
    for node_id, label in agent_node_labels.items()}
embedded_node_labels.update(global_node_labels)


# node colors
agent_process_colors = {
        'cobra':'tab:orange',
        'mass_deriver':'tab:orange',
        'volume_deriver':'tab:orange',
        'field_counts_deriver':'tab:brown',
        'biomass_adaptor':'tab:green',
        'bioscrape':'tab:green',
        'flux_adaptor':'tab:green',
        'dilution_rate_adaptor':'tab:green',
        'delta_counts_to_concs':'tab:green',
        'clock':'tab:gray',
        'strip_units':'tab:brown',
        'local_field':'tab:brown',
        'division':'tab:cyan',
        'divide_condition':'tab:cyan'}
agent_store_colors = {
        'flux_bounds':'tab:orange',
        'internal_counts':'tab:orange',
        'reactions':'tab:orange',
        'hidden_exchanges':'tab:orange',
        'cobra_external':'tab:orange',
        'rates':'tab:green',
        'species':'tab:green',
        'delta_species':'tab:green',
#         'delta\nconcs':'tab:green',
        'agents':'tab:cyan',
        'agents\n0':'tab:cyan',
        'boundary':'tab:brown',
        'boundary\nmass':'tab:brown',
        'boundary\ndivide':'tab:brown',
        'boundary\nexternal':'tab:brown',
        'boundary\nlocation':'tab:brown',
        'boundary\nexchanges':'tab:brown',
        'boundary\n(no units)':'tab:brown',
        'boundary\ntime':'tab:brown',
        'dimensions':'tab:brown',
#         'field_counts_deriver':'tab:brown',
        }
global_process_colors = {
        'diffusion':'tab:blue',
        'multibody':'tab:blue'}
global_store_colors = {
        'agents': 'tab:cyan',
        'agents\n0': 'tab:cyan',
        'fields':'tab:blue'}
embedded_process_colors = {
    'agents\n0\n'+node_id: color 
    for node_id, color in agent_process_colors.items()}
embedded_store_colors = {
    'agents\n0\n'+node_id: color 
    for node_id, color in agent_store_colors.items()}
embedded_process_colors.update(global_process_colors)
embedded_store_colors.update(global_store_colors)

embedded_custom_widths = {"agents\n0\n"+k:custom_widths[k] for k in custom_widths if k not in ["diffusion", "multibody"]}
embedded_custom_widths['diffusion'] = custom_widths['diffusion']
embedded_custom_widths['multibody'] = custom_widths['multibody']

config_bioscrape_cobra_topology = {
    'graph_format': 'hierarchy',
    'dashed_edges':True,
    'show_ports':False,
    'remove_nodes':[
        'boundary\ndimensions', 
        'boundary\nno_units',
        'boundary\ntime',
        'clock',
        'strip_units',
        'connect_external',
        'hidden_exchanges',
        'dimensions',
        'delta_concentrations',
        'boundary\nexchanges',
        
        'agents\n0\nboundary\ndimensions', 
        'agents\n0\nboundary\nno_units',
        'agents\n0\nboundary\ntime',
        'agents\n0\nclock',
        'agents\n0\nstrip_units',
        'agents\n0\nconnect_external',
        'agents\n0\nhidden_exchanges',
        'agents\n0\ndelta_concentrations',
        'agents\n0\nboundary\nexchanges'
    ],

}

config_single_cell_bioscrape_cobra_topology = copy.deepcopy(config_bioscrape_cobra_topology)
config_single_cell_bioscrape_cobra_topology['coordinates'] = agent_coordinates
config_single_cell_bioscrape_cobra_topology['node_labels'] = agent_node_labels
config_single_cell_bioscrape_cobra_topology['store_colors'] = agent_store_colors
config_single_cell_bioscrape_cobra_topology['process_colors'] = agent_process_colors
config_single_cell_bioscrape_cobra_topology['custom_widths'] = custom_widths

config_embedded_bioscrape_cobra_topology = copy.deepcopy(config_bioscrape_cobra_topology)
config_embedded_bioscrape_cobra_topology['coordinates'] = embedded_coordinates
config_embedded_bioscrape_cobra_topology['node_labels'] = embedded_node_labels
config_embedded_bioscrape_cobra_topology['process_colors'] = embedded_process_colors
config_embedded_bioscrape_cobra_topology['store_colors'] = embedded_store_colors
config_embedded_bioscrape_cobra_topology['custom_widths'] = embedded_custom_widths
config_embedded_bioscrape_cobra_topology['remove_nodes'].remove('dimensions')


import os
import glob
import setuptools
from distutils.core import setup

with open("README.md", 'r') as readme:
    long_description = readme.read()

setup(
    name='vivarium-notebooks',
    version='0.0.5',
    packages=[
        'bioscrape_cobra',
    ],
    author='Eran Agmon, William Poole',
    author_email='agmon.eran@gmail.com, wpoole@caltech.edu',
    url='https://github.com/vivarium-collective/vivarium-notebooks',
    license='MIT',
    entry_points={
        'console_scripts': []},
    short_description='Jupyter notebooks and supporting functions for the simulations '
                      'described in the paper "Vivarium: an interface and engine for '
                      'integrative multiscale modeling in computational biology"',
    long_description=long_description,
    long_description_content_type='text/markdown',
    package_data={},
    # package_data={
    #     # If any package contains *.xml files, include them:
    #     "": ["*.xml"],
    # },
    include_package_data=True,
    install_requires=[
        'vivarium-core>=0.3.5',
        'vivarium-bioscrape==0.0.0.7',
        'vivarium-cobra==0.0.18',
        'vivarium-multibody==0.0.13',
        'pytest',
        'tqdm',
        'jupyter'
    ],
)



"""
==============================================
Simulation helper functions for BioscrapeCOBRA
==============================================

Includes functions for configuring, running, and plotting all experiments reported in the paper:
    "Vivarium: an interface and engine for integrative multiscale modeling in computational biology"

These are used in the supplementary Jupyter notebook "Multi-Paradigm-Composites.ipynb".

This file also includes pre-configured experiments that can be triggered from the command line:
    * 1 : deterministic single cell
    * 2 : stochastic single cell
    * 3 : deterministic with cell division
    * 4 : stochastic with cell division
    * 5 : deterministic cells in spatial environment
    * 6 : stochastic cells in spatial environment

These can be triggered from the command line by entering the simulation number:
```
$ python bioscrape_cobra/simulate.py -sim_number
```

parallel and mongoDB database for saved output can be called with:
```
$ python bioscrape_cobra/simulate.py [-database, -d] [-parallel, -p]
```

see all options with:
```
python bioscrape_cobra/simulate.py -h
```

"""
import os
import argparse
import copy
import random
import time as clock
import pytest
from tqdm import tqdm

# vivarium imports
from vivarium.core.engine import Engine, timestamp
from vivarium.library.units import units
from vivarium.library.dict_utils import deep_merge
from vivarium.core.composition import simulate_composer, simulate_composite
from vivarium.core.composer import Composite
from vivarium.library.units import remove_units
from vivarium.core.serialize import deserialize_value

# vivarium-multibody imports
from vivarium_multibody.composites.lattice import (
    Lattice, make_lattice_config)
from vivarium_multibody.composites.grow_divide import GrowDivide
from vivarium_multibody.processes.diffusion_field import DiffusionField

# cobra imports
from vivarium_cobra.processes.cobra_fba import (
    COBRA_FBA, 
    get_iAF1260b_config, 
)
from vivarium_cobra.composites.cobra_composite import CobraComposite
from vivarium_cobra.library.lattice_utils import get_bin_volume

# Bioscrape imports
from vivarium_bioscrape.processes.bioscrape import Bioscrape

# local import
from bioscrape_cobra.bioscrape_cobra_stochastic import (
    BioscrapeCOBRAstochastic, SBML_FILE_STOCHASTIC, COBRA_TIMESTEP, BIOSCRAPE_TIMESTEP)
from bioscrape_cobra.bioscrape_cobra_stochastic import GLUCOSE_EXTERNAL, LACTOSE_EXTERNAL
from bioscrape_cobra.bioscrape_cobra_deterministic import BioscrapeCOBRAdeterministic, SBML_FILE_DETERMINISTIC

# plotting
from vivarium.plots.topology import plot_topology
from vivarium.plots.simulation_output import _save_fig_to_dir as save_fig_to_dir
from bioscrape_cobra.plot import (
    plot_multigen, plot_single, plot_fields_tags, plot_fields_snapshots, config_embedded_bioscrape_cobra_topology)

# get the sbml files at the correct path
dirname = os.path.dirname(__file__)
sbml_deterministic_file = os.path.join(dirname, SBML_FILE_DETERMINISTIC)
sbml_stochastic_file = os.path.join(dirname, SBML_FILE_STOCHASTIC)


# default environment variables, which can be varied by the function `simulate_bioscrape_cobra`
DEFAULT_DIVIDE_THRESHOLD = 2000 * units.fg
INITIAL_GLC = 10  # mmolar
INITIAL_LAC = 20  # mmolar
BOUNDS = [20, 20]
NBINS = [10, 10]
DEPTH = 2

# plotting definitions
plot_variables_list = [
    ('species', 'rna_M'),
    ('species', 'monomer_betaGal'),
    ('species', 'protein_betaGal'),
    ('species', 'protein_Lactose_Permease'),
    ('species', GLUCOSE_EXTERNAL),
    ('species', LACTOSE_EXTERNAL),
    ('flux_bounds', 'EX_glc__D_e'),
    ('flux_bounds', 'EX_lac__D_e'),
    ('boundary', 'no_units', 'biomass'),
]
plot_variables_list_deterministic = [
    ('rates', 'k_dilution__',)]
plot_variables_list_deterministic.extend(plot_variables_list)
plot_variables_list_stochastic = []
plot_variables_list_stochastic.extend(plot_variables_list)


# Simulates Cobra Model on its Own
def simulate_cobra(
    total_time=100,
    initial_state=None,
):
    # get the configuration for the iAF1260b BiGG model
    iAF1260b_config = get_iAF1260b_config()
    iAF1260b_config.update({'time_step': COBRA_TIMESTEP})

    iAF1260b_config['name'] = 'cobra' #Rename the process
    dynamic_fba = COBRA_FBA(iAF1260b_config)

    cobra_composite = Composite({
        'processes': dynamic_fba, 
        'topology': dynamic_fba.generate_topology()
        })

    # get the initial state
    if initial_state is None:
        initial_state = dynamic_fba.initial_state({})

    # run simulation
    cobra_sim_settings = {
        'initial_state': initial_state,
        'total_time': total_time}
    cobra_timeseries = simulate_composer(dynamic_fba, cobra_sim_settings)

    return cobra_timeseries, dynamic_fba


# Simulates Cobra Model in a Composite with some Derivers
def simulate_cobra_composite(
    total_time=100,
    initial_state=None,
):

    # get the configuration for the iAF1260b BiGG model
    iAF1260b_config = get_iAF1260b_config()
    iAF1260b_config.update({'time_step': COBRA_TIMESTEP})

    #Place the Process into a Composite level dictionary
    composite_config = {'cobra': iAF1260b_config}
    cobra_composite = CobraComposite(composite_config)

    # get the initial state
    if initial_state is None:
        initial_state = cobra_composite.initial_state({})

    # run simulation
    cobra_sim_settings = {
        'initial_state': initial_state,
        'total_time': total_time}
    cobra_timeseries = simulate_composer(cobra_composite, cobra_sim_settings)

    return cobra_timeseries, cobra_composite


def simulate_bioscrape(
        stochastic=False,
        initial_glucose=None,
        initial_lactose=None,
        initial_state=None,
        total_time=100,
        initial_volume=1.0,
        sbml_file=None,
):
    
    #create configs
    if not stochastic:
        bioscrape_config = {
            'sbml_file': sbml_file or sbml_deterministic_file,
            'stochastic': False,
            'initial_volume': initial_volume,
            'internal_dt': BIOSCRAPE_TIMESTEP/100,
            'time_step': BIOSCRAPE_TIMESTEP}
    else:
        bioscrape_config = {
            'sbml_file': sbml_file or sbml_stochastic_file,
            'stochastic': True,
            'safe_mode': False,
            'initial_volume': initial_volume,
            'internal_dt': BIOSCRAPE_TIMESTEP/100,
            'time_step': BIOSCRAPE_TIMESTEP
        }
    
    bioscrape_process = Bioscrape(bioscrape_config)


    #Create an Empty Composite for Plotting Purposes
    bioscrape_composite = Composite({
        'processes': bioscrape_process.generate_processes(), 
        'topology': bioscrape_process.generate_topology()
        })

    # get the initial state
    if initial_state is None:
        initial_state = bioscrape_process.initial_state({})

    if initial_glucose is not None:
        initial_state['species']['Glucose_external'] = initial_glucose

    if initial_lactose is not None:
        initial_state['species']['Lactose_external'] = initial_lactose

    # run simulation
    bioscrape_sim_settings = {
        'initial_state': initial_state,
        'total_time': total_time
        }

    bioscrape_timeseries = simulate_composer(bioscrape_process, bioscrape_sim_settings)

    return bioscrape_timeseries, bioscrape_composite


# Simulate a system of cells that grow and divide in a well mixed spatial environment
def simulate_grow_divide(
        total_time=100,
        growth_rate=0.03,
        initial_state=None,
        growth_noise=10**-6
):
    # config
    growth_config = {
        'default_growth_rate': growth_rate,
        'default_growth_noise': growth_noise,
        'agents_path': ('agents',)}
    grow_divide_composer = GrowDivide({
        'agent_id': '0', 'growth' : growth_config})

    if initial_state is None:
        initial_state = {
            'agents': {
                '0': {
                    'global': {
                        'mass': 1000 * units.femtogram}
                }}}

    grow_divide_sim_settings = {
        'total_time': total_time,
        'initial_state': initial_state,
        'outer_path': ('agents', '0'),
        'return_raw_data': True}
    grow_divide_composite = grow_divide_composer.generate(path=('agents', '0'))

    grow_divide_data = simulate_composite(grow_divide_composite, grow_divide_sim_settings)
    grow_divide_data = deserialize_value(grow_divide_data)
    grow_divide_data = remove_units(grow_divide_data)

    #returns the data, the initial composite, and the final composite
    return grow_divide_data, grow_divide_composer.generate(path=('agents', '0')), grow_divide_composite


def simulate_diffusion(
        total_time=100,
        diffusion_rate=0.001,
        initial_state={},
        bins=[10, 10],
        bounds=[10, 10]
):
    # configure
    

    config = {
        'n_bins':bins,
        'bounds':bounds,
        'diffusion':diffusion_rate,
        'initial_state':{'glc':initial_state}
    }

    diffusion_process = DiffusionField(config)

    diffusion_composer = Composite({
        'processes': diffusion_process.generate_processes(), 
        'topology': diffusion_process.generate_topology()
        })

    diffusion_state = diffusion_process.initial_state({})

    diffusion_sim_settings = {
        'total_time': total_time,
        'return_raw_data': True,
        'initial_state': diffusion_state,
    }

    diffusion_data = simulate_composer(diffusion_process, diffusion_sim_settings)

    #Add empty agents to reuse plotting functionality
    for t in diffusion_data:
        diffusion_data[t]['agents'] = {}

    return diffusion_data, diffusion_composer


def get_lattice_grow_divide_composite(
        diffusion_rate=0.001,
        initial_concentration={},
        bins=[10, 10],
        bounds=[10, 10],
        growth_rate=0.03,
        growth_noise=10**-6,
        depth=10
):
    lattice_config = make_lattice_config(
            bounds=bounds,
            n_bins=bins,
            concentrations={'glc':initial_concentration},
            diffusion=diffusion_rate,
            depth=depth)

    lattice_composer = Lattice(lattice_config)
    lattice_composite = lattice_composer.generate()


    growth_config = {'default_growth_rate': growth_rate, 'default_growth_noise': growth_noise}
    grow_divide_composer = GrowDivide({'agent_id': '0', 'growth': growth_config})

    agent_id = '0'
    lattice_grow_divide_composite = grow_divide_composer.generate(path=('agents', agent_id))
    lattice_grow_divide_composite.merge(composite=lattice_composite)

    return lattice_grow_divide_composite


def simulate_grow_divide_lattice(
        lattice_grow_divide_composite,
        total_time=100,
        initial_state=None,
):

    agent_id = '0'
    if initial_state is None:
        initial_state = {
            'agents': {
                agent_id: {
                    'global': {
                        'mass': 1000 * units.femtogram}
                }}}

    sim_settings = {
        'total_time': total_time,
        'return_raw_data': True,
        'initial_state': initial_state,
        'return_raw_data': True}
    lattice_grow_divide_data = simulate_composite(
        lattice_grow_divide_composite, sim_settings)

    return lattice_grow_divide_data


# helper functions
def get_bioscrape_cobra_config(
        spatial=False,
        division=False,
        divide_threshold=DEFAULT_DIVIDE_THRESHOLD,
        external_volume=None,
        sbml_file=None,
        parallel=False,
):
    """ create a generic config dict for bioscrape_cobra composers """
    config = {
        '_parallel': parallel,
        **({'local_fields': {'bin_volume': external_volume}}
           if external_volume is not None else {}),
        **({'sbml_file': sbml_file} if sbml_file is not None else {}),
    }

    if spatial or division:
        config.update({
            'divide_on': True,
            'agents_path': ('..', '..', 'agents',),
            'fields_path': ('..', '..', 'fields',),
            'dimensions_path': ('..', '..', 'dimensions',),
            'divide_condition': {
                'threshold': divide_threshold},
        })

    if spatial:
        config.update({
            'fields_on': True})

    return config


def get_initial_state(initial_states, n=None):
    """ return an initial state from the initial_states argument """
    state_copy = copy.deepcopy(initial_states)
    if isinstance(state_copy, list):
        if n is not None and len(state_copy) > n:
            return state_copy[n]
        else:
            return random.choice(state_copy)
    elif isinstance(state_copy, dict):
        return state_copy
    else:
        return {}


def simulate_bioscrape_cobra(
        division=False,
        stochastic=False,
        initial_glucose=1e1,
        initial_lactose=1e1,
        initial_agent_states=None,
        bounds=None,
        n_bins=None,
        depth=DEPTH,
        diffusion_rate=1e-1,
        jitter_force=1e-4,
        divide_threshold=2000 * units.fg,
        spatial=False,
        external_volume=None,
        n_agents=1,
        halt_threshold=100,
        total_time=100,
        sbml_file=None,
        emitter='timeseries',
        output_type=None,
        parallel=False,
):
    """ Main simulation function for BioscrapeCOBRA

    Args:
        * division (bool): sets whether the agents divides
        * stochastic (bool): load the stochastic lac operon model
        * initial_glucose (float): initial external glucose concentration
        * initial_lactose: (float): initial external initial_lactose concentration
        * initial_agent_states (dict): set initial state values
        * bounds (list): size of the environment [x, y] in microns
        * n_bins (list): number of bins in the [x, y] dimensions
        * depth (float): depth of the environment in microns
        * diffusion_rate (float): diffusion rate constant for all molecules, micron^s/sec.
        * divide_threshold (float): mass at which cells divide, in fg
        * spatial (bool): use spatial environment
        * external_volume (float): volume of external bin, if non-spatial environment
        * n_agents (int): number of initial agents in environment
        * halt_threshold (int): number of agents at which simulations will terminate
        * total_time (float): total simulation time, in seconds
        * sbml_file (str): the file for the Bioscrape process. Uses default if None.
        * emitter (str): type of emitter, 'timeseries' or 'database'.
        * output_type (str): 'timeseries' or 'unitless'. If None, return experiment instance
        * parallel (bool): run processes in parallel, useful for large compute machines
    """

    if n_bins is None:
        n_bins = NBINS
    if bounds is None:
        bounds = BOUNDS
    if sbml_file is None:
        if stochastic:
            sbml_file = sbml_stochastic_file
        else:
            sbml_file = sbml_deterministic_file

    # get the bin volume based upon the lattice
    bin_volume = (external_volume or get_bin_volume(n_bins, bounds, depth)) * units.L
    agent_state = {
        # set initial flux values based on COBRA defaults.
        'flux_bounds': {
            'EX_lac__D_e': 0.0,
            'EX_glc__D_e': 0.099195
        },
        # field_counts_deriver needs the bin volume
        'boundary': {
            'bin_volume': bin_volume,
            **({'external': {
                GLUCOSE_EXTERNAL: initial_glucose,
                LACTOSE_EXTERNAL: initial_lactose}
               } if spatial else {}),
        },
    }

    # make the BioscrapeCOBRA config
    biocobra_config = get_bioscrape_cobra_config(
        spatial=spatial,
        division=division,
        divide_threshold=divide_threshold,
        external_volume=bin_volume,
        sbml_file=sbml_file,
        parallel=parallel)

    # get the BioscrapeCOBRA composer -- either stochastic or deterministic
    if stochastic:
        biocobra_composer = BioscrapeCOBRAstochastic(biocobra_config)
    else:
        biocobra_composer = BioscrapeCOBRAdeterministic(biocobra_config)

    # initialize the composite
    biocobra_composite = Composite({
        'processes': {},
        'topology': {}})

    if spatial:
        # make n_agents dividing agents and combine them with a Lattice environment

        agents_initial = {'agents': {}}
        for n in range(n_agents):
            agent_id = str(n)
            config = {'agent_id': agent_id}
            agent = biocobra_composer.generate(config=config)
            biocobra_composite.merge(
                composite=agent,
                path=('agents', agent_id))

            # initial state for agent
            agents_initial['agents'][agent_id] = get_initial_state(
                initial_agent_states, n=n)
            agents_initial['agents'][agent_id] = deep_merge(
                agents_initial['agents'][agent_id], agent_state)

        # create a second initial composite for plotting
        initial_composite = biocobra_composer.generate(
            path=('agents', '0'),
            config={'agent_id': '0'})

        # make a lattice composite for the environment
        field_concentrations = {
            GLUCOSE_EXTERNAL: initial_glucose,
            LACTOSE_EXTERNAL: initial_lactose}
        lattice_config = make_lattice_config(
            bounds=bounds,
            n_bins=n_bins,
            depth=depth,
            concentrations=field_concentrations,
            diffusion=diffusion_rate,
            jitter_force=jitter_force,
            time_step=min(COBRA_TIMESTEP, BIOSCRAPE_TIMESTEP), #/2
        )
        lattice_config['multibody']['_parallel'] = parallel
        lattice_composer = Lattice(lattice_config)

        # merge bioscrapeCOBRA composite with lattice
        biocobra_composite.merge(composite=lattice_composer.generate())
        initial_composite.merge(composite=lattice_composer.generate())

        # get initial state from the composite and merge declared initial
        state = biocobra_composite.initial_state()
        initial_state_full = deep_merge(state, agents_initial)

    elif division:
        # make n_agents dividing agents, without an explicit environment

        # division requires the agent to be embedded in a hierarchy
        # make the bioscrapeCOBRA composite under the path ('agents', agent_id)
        agents_initial = {'agents': {}}
        for n in range(n_agents):
            agent_id = str(n)
            config = {'agent_id': agent_id}
            agent = biocobra_composer.generate(config=config)
            biocobra_composite.merge(
                composite=agent,
                path=('agents', agent_id))

            # initial state for agent
            agents_initial['agents'][agent_id] = get_initial_state(initial_agent_states, n=n)
            agents_initial['agents'][agent_id] = deep_merge(
                agents_initial['agents'][agent_id], agent_state)

        # create a second initial composite for plotting
        initial_composite = biocobra_composer.generate(
            path=('agents', '0'),
            config={'agent_id': '0'})

        # get initial state from the composite and merge declared initial
        state = biocobra_composite.initial_state()
        initial_state_full = deep_merge(state, agents_initial)

        # simple, 1D field
        initial_state_full['fields'] = {
                GLUCOSE_EXTERNAL: initial_glucose,
                LACTOSE_EXTERNAL: initial_lactose}

    else:
        # single agent without division

        # make the composite
        agent = biocobra_composer.generate()
        biocobra_composite.merge(
            composite=agent,
            path=())

        # create a second initial composite for plotting
        initial_composite = biocobra_composer.generate()

        # get initial state from the composite and merge declared initial
        agents_initial = biocobra_composite.initial_state()
        initial_declared = get_initial_state(initial_agent_states)
        agents_initial = deep_merge(agents_initial, initial_declared)
        initial_state_full = deep_merge(agents_initial, agent_state)

        # simple, 1D field
        initial_state_full['fields'] = {
                GLUCOSE_EXTERNAL: initial_glucose,
                LACTOSE_EXTERNAL: initial_lactose}

    # make the experiment
    experiment_id = (f"{'stochastic' if stochastic else 'deterministic'}"
                     f"{'_division' if division else ''}"
                     f"{'_spatial' if spatial else ''}"
                     f"_{timestamp()}")
    experiment_config = {
        'processes': biocobra_composite.processes,
        'topology': biocobra_composite.topology,
        'initial_state': initial_state_full,
        'display_info': False,
        'experiment_id': experiment_id,
        'emit_step': max(BIOSCRAPE_TIMESTEP, COBRA_TIMESTEP),
        'emitter': {'type': emitter}}
    print(f'Initializing experiment {experiment_id}')
    biocobra_experiment = Engine(**experiment_config)

    # run the experiment
    clock_start = clock.time()
    if division:  # terminate upon reaching total_time or halt_threshold
        sim_step = max(BIOSCRAPE_TIMESTEP, COBRA_TIMESTEP) * 10
        for _ in tqdm(range(0, total_time, sim_step)):
            n_agents = len(biocobra_experiment.state.get_value()['agents'])
            if n_agents < halt_threshold:
                biocobra_experiment.update(sim_step)
    else:
        biocobra_experiment.update(total_time)

    # print runtime and finalize
    clock_finish = clock.time() - clock_start
    print(f'Completed in {clock_finish:.2f} seconds')
    biocobra_experiment.end()

    # retrieve the data
    if output_type == 'timeseries':
        return biocobra_experiment.emitter.get_timeseries(), initial_composite
    if output_type == 'unitless':
        return biocobra_experiment.emitter.get_data_unitless(), initial_composite
    return biocobra_experiment, initial_composite


# plot topology
def plot_full_topology(out_dir='out'):

    dirname = os.path.dirname(__file__)
    sbml_file = os.path.join(dirname, SBML_FILE_STOCHASTIC)

    # get the stochastic bioscrape_cobra composite
    output, composite = simulate_bioscrape_cobra(
        stochastic=True,
        division=True,
        spatial=True,
        total_time=0,
        sbml_file=sbml_file,
        output_type='unitless')

    plot_topology(
        composite,
        settings=config_embedded_bioscrape_cobra_topology,
        out_dir=out_dir,
        filename='bioscrape_cobra_stochastic_lattice_topology.pdf')


# tests
def test_deterministic(
        emitter=None,
        sbml_deterministic=None,
        total_time=200,
        out_dir=None,
):
    if not sbml_deterministic:
        dirname = os.path.dirname(__file__)
        sbml_deterministic = os.path.join(dirname, SBML_FILE_DETERMINISTIC)

    output, comp0 = simulate_bioscrape_cobra(
        initial_glucose=1e1,
        initial_lactose=2e1,
        external_volume=1e-12,
        total_time=total_time,
        emitter=emitter,
        sbml_file=sbml_deterministic,
        output_type='timeseries')

    if out_dir:
        plot_single(
            output,
            variables=plot_variables_list_deterministic,
            out_dir=os.path.join(out_dir, 'deterministic'),
            filename='variables')


@pytest.mark.slow
def test_stochastic(
        emitter=None,
        sbml_stochastic=None,
        total_time=500,
        out_dir=None,
):
    if not sbml_stochastic:
        dirname = os.path.dirname(__file__)
        sbml_stochastic = os.path.join(dirname, SBML_FILE_STOCHASTIC)

    initial_agent_state = {
        'rates': {
            'k_leak': 0.6,
        },
        'species': {
            'monomer_betaGal': 100,
            'protein_betaGal': 100,
            'protein_Lactose_Permease': 100}}

    output, comp0 = simulate_bioscrape_cobra(
        stochastic=True,
        initial_glucose=1e1,
        initial_lactose=2e1,
        initial_agent_states=initial_agent_state,
        external_volume=1e-14,
        total_time=total_time,
        emitter=emitter,
        sbml_file=sbml_stochastic,
        output_type='timeseries')

    if out_dir:
        plot_single(
            output,
            variables=plot_variables_list_stochastic,
            out_dir=os.path.join(out_dir, 'stochastic'),
            filename='variables')


def test_deterministic_divide(
        emitter=None,
        sbml_deterministic=None,
        total_time=200,
        out_dir=None,
):
    if not sbml_deterministic:
        dirname = os.path.dirname(__file__)
        sbml_deterministic = os.path.join(dirname, SBML_FILE_DETERMINISTIC)

    initial_agent_states = [
        {'species': {
            'monomer_betaGal': 0.0,
            'protein_betaGal': 0.0,
            'protein_Lactose_Permease': 0.0}},
        {'species': {
            'monomer_betaGal': 0.0,
            'protein_betaGal': 0.1,
            'protein_Lactose_Permease': 0.1}}]

    output, comp0 = simulate_bioscrape_cobra(
        n_agents=2,
        initial_agent_states=initial_agent_states,
        division=True,
        initial_glucose=1e1,
        initial_lactose=2e1,
        external_volume=1e-12,
        total_time=total_time,
        emitter=emitter,
        sbml_file=sbml_deterministic,
        output_type='unitless')

    if out_dir:
        var_list = copy.deepcopy(plot_variables_list_deterministic)
        plot_multigen(
            output,
            variables=var_list,
            out_dir=os.path.join(out_dir, 'deterministic_divide'),
            filename='division_multigen')


@pytest.mark.slow
def test_stochastic_divide(
        emitter=None,
        sbml_stochastic=None,
        total_time=200,
        out_dir=None,
):
    if not sbml_stochastic:
        dirname = os.path.dirname(__file__)
        sbml_stochastic = os.path.join(dirname, SBML_FILE_STOCHASTIC)

    initial_agent_states = [{
            'rates': {
                'LacPermease_vmax': 3580.0,  # 35.8
                'k_leak': 0.5,
            },
        }]

    output, comp0 = simulate_bioscrape_cobra(
        n_agents=1,
        stochastic=True,
        division=True,
        initial_glucose=1e1,  # mM
        initial_lactose=2e1,  # mM
        initial_agent_states=initial_agent_states,
        total_time=total_time,
        external_volume=5e-14,
        emitter=emitter,
        sbml_file=sbml_stochastic,
        output_type='unitless')

    if out_dir:
        # plot
        var_list = copy.deepcopy(plot_variables_list_stochastic)
        var_list.extend([
            ('boundary', 'mass'),
            ('boundary', 'volume')])
        plot_multigen(
            output,
            variables=var_list,
            out_dir=os.path.join(out_dir, 'stochastic_divide'),
            filename='division_multigen')


def test_deterministic_spatial(
        emitter=None,
        sbml_deterministic=None,
        total_time=200,
        out_dir=None,
):
    if not sbml_deterministic:
        dirname = os.path.dirname(__file__)
        sbml_deterministic = os.path.join(dirname, SBML_FILE_DETERMINISTIC)

    bounds = [30, 30]
    n_bins = [30, 30]
    depth = 1

    initial_agent_states = [
        {'rates': {
            'k_leak': 0.5,
        }}
    ]

    output, comp0 = simulate_bioscrape_cobra(
        initial_agent_states=initial_agent_states,
        division=True,
        spatial=True,
        initial_glucose=1e1,
        initial_lactose=2e1,
        bounds=bounds,
        n_bins=n_bins,
        depth=depth,
        total_time=total_time,
        emitter=emitter,
        sbml_file=sbml_deterministic,
        output_type='unitless')

    if out_dir:
        # plot
        deterministic_spatial_out_dir = os.path.join(out_dir, 'deterministic_spatial')
        plot_multigen(
            output,
            out_dir=deterministic_spatial_out_dir,
            filename='spatial_multigen')

        plot_fields_snapshots(
            output,
            bounds=bounds,
            include_fields=[GLUCOSE_EXTERNAL, LACTOSE_EXTERNAL],
            out_dir=deterministic_spatial_out_dir,
            filename='spatial_snapshots')

        plot_fields_tags(
            output,
            bounds=bounds,
            tagged_molecules=[('species', 'protein_Lactose_Permease',)],
            out_dir=deterministic_spatial_out_dir,
            filename='spatial_tags')


@pytest.mark.slow
def test_stochastic_spatial(
    emitter=None,
    parallel=False,
    sbml_stochastic=None,
    total_time=200,
    out_dir=None,
):
    if not sbml_stochastic:
        dirname = os.path.dirname(__file__)
        sbml_stochastic = os.path.join(dirname, SBML_FILE_STOCHASTIC)

    bounds = [30, 30]
    n_bins = [30, 30]
    depth = 0.5

    initial_agent_states = [
        {'rates': {
            'k_leak': 0.005  # less leak -> less spontanteous expression
        }}
    ]

    output, comp0 = simulate_bioscrape_cobra(
        initial_agent_states=initial_agent_states,
        stochastic=True,
        division=True,
        spatial=True,
        initial_glucose=1e1,
        initial_lactose=5e1,
        depth=depth,
        diffusion_rate=2e-2,
        jitter_force=1e-5,
        bounds=bounds,
        n_bins=n_bins,
        halt_threshold=200,
        total_time=total_time,
        emitter=emitter,
        sbml_file=sbml_stochastic,
        parallel=parallel,
        output_type='unitless')

    if out_dir:
        # plot
        stochastic_spatial_out_dir = os.path.join(out_dir, 'stochastic_spatial')
        plot_multigen(
            output,
            out_dir=stochastic_spatial_out_dir,
            filename='spatial_multigen')

        plot_fields_snapshots(
            output,
            bounds=bounds,
            include_fields=[GLUCOSE_EXTERNAL, LACTOSE_EXTERNAL],
            out_dir=stochastic_spatial_out_dir,
            filename='spatial_snapshots')

        plot_fields_tags(
            output,
            bounds=bounds,
            convert_to_concs=False,
            tagged_molecules=[('species', 'protein_Lactose_Permease',)],
            out_dir=stochastic_spatial_out_dir,
            filename='spatial_tags')


def test_bioscrape_alone(
        sbml_deterministic=None,
        sbml_stochastic=None,
        out_dir=None,
):
    if not sbml_deterministic:
        dirname = os.path.dirname(__file__)
        sbml_deterministic = os.path.join(dirname, SBML_FILE_DETERMINISTIC)
    if not sbml_stochastic:
        dirname = os.path.dirname(__file__)
        sbml_stochastic = os.path.join(dirname, SBML_FILE_STOCHASTIC)


    # Simulate the Lac Operon CRN Deterministically
    bioscrape_timeseries_det, bioscrape_composite_det = simulate_bioscrape(
        total_time=20000,
        initial_glucose=10,
        initial_lactose=20,
        sbml_file=sbml_deterministic)
    # Simulate the Lac Operon CRN Stochastically
    bioscrape_timeseries_stoch, bioscrape_composite_stoch = simulate_bioscrape(
        total_time=20000,
        initial_glucose=10 * 6.22 * 10 ** 6,
        initial_lactose=20 * 6.22 * 10 ** 6,
        stochastic=True,
        sbml_file=sbml_stochastic)

    if out_dir:
        # Plot the CRN Trajectories
        species_to_plot_det = [
            {'variable': ('species', 'Glucose_external'), 'display': 'external glucose (mM)'},
            {'variable': ('species', 'Lactose_external'), 'display': 'external lactose (mM)'},
            {'variable': ('species', 'rna_M'), 'display': 'lac operon RNA (mM)'},
            {'variable': ('species', 'protein_betaGal'), 'display': r'$\beta$-galactosidase (mM)'},
            {'variable': ('species', 'protein_Lactose_Permease'), 'display': 'lactose permease (mM)'}
        ]
        species_to_plot_stoch = [
            {'variable': ('species', 'Glucose_external'), 'display': 'external glucose  (counts)'},
            {'variable': ('species', 'Lactose_external'), 'display': 'external lactose (counts)'},
            {'variable': ('species', 'rna_M'), 'display': 'lac operon RNA (counts)'},
            {'variable': ('species', 'protein_betaGal'), 'display': r'$\beta$-galactosidase (counts)'},
            {'variable': ('species', 'protein_Lactose_Permease'), 'display': 'lactose permease (counts)'}
        ]

        fig_timeseries = plot_single(
            bioscrape_timeseries_det,
            variables=species_to_plot_det)
        save_fig_to_dir(
            fig_timeseries,
            filename='bioscrape_deterministic_output.pdf',
            out_dir=out_dir)

        fig_timeseries = plot_single(
            bioscrape_timeseries_stoch,
            variables=species_to_plot_stoch)
        save_fig_to_dir(
            fig_timeseries,
            filename='bioscrape_stochastic_output.pdf',
            out_dir=out_dir)


def main():
    out_dir = os.path.join('out', 'bioscrape_cobra')
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    parser = argparse.ArgumentParser(description='bioscrape_cobra')
    parser.add_argument('-database', '-d', action='store_true', default=False, help='emit to database')
    parser.add_argument('-parallel', '-p', action='store_true', default=False, help='run parallel processes')
    parser.add_argument('--deterministic', '-1', action='store_true', default=False)
    parser.add_argument('--stochastic', '-2', action='store_true', default=False)
    parser.add_argument('--deterministic_divide', '-3', action='store_true', default=False)
    parser.add_argument('--stochastic_divide', '-4', action='store_true', default=False)
    parser.add_argument('--deterministic_spatial', '-5', action='store_true', default=False)
    parser.add_argument('--stochastic_spatial', '-6', action='store_true', default=False)
    parser.add_argument('--topology', '-t', action='store_true', default=False)
    parser.add_argument('--bioscrape-alone', '-b', action='store_true', default=False)

    args = parser.parse_args()

    # emitter type
    emitter = 'database' if args.database else 'timeseries'
    parallel = True if args.parallel else False

    if args.deterministic:
        test_deterministic(
            emitter=emitter,
            sbml_deterministic=sbml_deterministic_file,
            total_time=12000,
            out_dir=out_dir
        )

    if args.stochastic:
        test_stochastic(
            emitter=emitter,
            sbml_stochastic=sbml_stochastic_file,
            total_time=4000,
            out_dir=out_dir,
        )

    if args.deterministic_divide:
        test_deterministic_divide(
            emitter=emitter,
            sbml_deterministic=sbml_deterministic_file,
            total_time=3000,
            out_dir=out_dir
        )

    if args.stochastic_divide:
        test_stochastic_divide(
            emitter=emitter,
            sbml_stochastic=sbml_stochastic_file,
            total_time=4000,
            out_dir=out_dir,
        )

    if args.deterministic_spatial:
        test_deterministic_spatial(
            emitter=emitter,
            sbml_deterministic=sbml_deterministic_file,
            total_time=7200,
            out_dir=out_dir,
        )

    if args.stochastic_spatial:
        test_stochastic_spatial(
            emitter=emitter,
            parallel=parallel,
            sbml_stochastic=sbml_stochastic_file,
            total_time=50400,
            out_dir=out_dir,
        )

    if args.topology:
        plot_full_topology(out_dir=out_dir)

    if args.bioscrape_alone:
        test_bioscrape_alone()


if __name__ == '__main__':
    main()

